/*
 * Copyright 2011 NORDUnet A/S.  All rights reserved.
 * Copyright 2011 Red Hat, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "asn1/PA-OTP-ENC-REQUEST.h"
#include "asn1/PA-OTP-CHALLENGE.h"
#include "asn1/PA-OTP-REQUEST.h"

#define CLIENT
#include "asn1c_utils.h"
#include "codec.h"
#include "internal.h"
#include "memory.h"

#include <krb5/preauth_plugin.h>

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#define COPY_FIELD(req, ti, field, type) \
  asn_DEF_ ## type.free_struct(&asn_DEF_ ## type, req->field, 0); \
  req->field = ti->field; ti->field = NULL

static krb5_preauthtype authhub_client_supported_pa_types[] =
  { KRB5_PADATA_OTP_CHALLENGE, 0 };

static int
authhub_client_init(krb5_context context, krb5_clpreauth_moddata *moddata_out)
{
  PA_OTP_REQUEST_t *req = new0(PA_OTP_REQUEST_t);
  if (req) {
    req->flags.buf = new0(uint32_t);
    req->flags.size = sizeof(uint32_t);
    if (!req->flags.buf) {
      free(req);
      return ENOMEM;
    }
  } else
    return ENOMEM;

  *moddata_out = (krb5_clpreauth_moddata) req;
  return 0;
}

static void
authhub_client_fini(krb5_context context, krb5_clpreauth_moddata moddata)
{
  asn_DEF_PA_OTP_REQUEST.free_struct(&asn_DEF_PA_OTP_REQUEST, moddata, 0);
}

static int
authhub_client_get_flags(krb5_context context, krb5_preauthtype pa_type)
{
  return PA_REAL;
}

static krb5_error_code
encrypt_nonce(krb5_context ctx, krb5_keyblock *key,
    const PA_OTP_CHALLENGE_t *chl, PA_OTP_REQUEST_t *req)
{
  PA_OTP_ENC_REQUEST_t er;
  krb5_error_code retval;
  krb5_enc_data encdata;
  krb5_data data;
  size_t size;

  /* Encode the nonce */
  er.nonce = chl->nonce;
  if (!int_der_encode(&asn_DEF_PA_OTP_ENC_REQUEST, &er, &data.data,
      &data.length))
    return ENOMEM;

  /* Get the size of the encrypted data */
  retval = krb5_c_encrypt_length(ctx, key->enctype, data.length, &size);
  if (retval != 0) {
    DEBUGMSG(retval, "krb5_c_encrypt_length failed.");
    free(data.data);
    return retval;
  }

  /* Allocate a buffer to put encrypted data in */
  encdata.ciphertext.length = size;
  encdata.ciphertext.data = malloc0(size);
  if (encdata.ciphertext.data == NULL) {
    free(data.data);
    return ENOMEM;
  }

  /* Do the encryption */
  retval = krb5_c_encrypt(ctx, key, KRB5_KEYUSAGE_PA_OTP_REQUEST,
                          NULL, &data, &encdata);
  free(data.data);
  if (retval != 0) {
    DEBUGMSG(retval, "Failed to encrypt nonce.");
    free(encdata.ciphertext.data);
    return retval;
  }

  /* Set the encrypted data */
  asn_DEF_OCTET_STRING.free_struct(&asn_DEF_OCTET_STRING,
                                   &req->encData.cipher, 1);
  req->encData.cipher.buf = (unsigned char*) encdata.ciphertext.data;
  req->encData.cipher.size = encdata.ciphertext.length;
  req->encData.etype = encdata.enctype;
  req->encData.kvno = new0(UInt32_t);
  if (asn_long2INTEGER(req->encData.kvno, encdata.kvno) < 0) {
    asn_DEF_OCTET_STRING.free_struct(&asn_DEF_OCTET_STRING,
                                     &req->encData.cipher, 1);
    memset(&req->encData, 0, sizeof(req->encData));
    return ENOMEM;
  }

  return 0;
}

static void
on_constraint_fail(void *app_key, struct asn_TYPE_descriptor_s *failed_type,
                   const void *failed_ptr, const char *errmsg_format, ...)
{
  va_list ap;
  char *tmp;

  va_start(ap, errmsg_format);
  if (vasprintf(&tmp, errmsg_format, ap) < 0)
    return;
  va_end(ap);

  DEBUGMSG(EINVAL, "Constraint Failed: %s", tmp);
  free(tmp);
}

static int
isalphanumeric(int c)
{
  if (c >= 'A' && c <= 'Z')
    return 1;
  if (c >= 'a' && c <= 'z')
    return 1;
  if (c >= '0' && c <= '9')
    return 1;
  return 0;
}

static int
isbinary(int c)
{
  return 1;
}

static int
isbase64(int c)
{
  return isalphanumeric(c) || c == '+' || c == '/';
}

static bool
token_matches_tokeninfo(const char *token, OTP_TOKENINFO_t *ti)
{
  int (*table[])(int c) =
      { isdigit, isxdigit, isalphanumeric, isbinary, isbase64 };

  if (!token || !ti)
    return false;

  if (ti->otp_length && strlen(token) != *ti->otp_length)
    return false;

  if (ti->otp_format) {
    long int format;

    if (asn_INTEGER2long(ti->otp_format, &format) < 0
        || format < OTPFormat_decimal
        || format > OTPFormat_base64)
      return false;

    while (*token)
      if (!(*table[format])(*token++))
    return false;
  }

  return true;
}

static void
remove_tokeninfo(PA_OTP_CHALLENGE_t *chl, int i)
{
  asn_DEF_OTP_TOKENINFO.free_struct(&asn_DEF_OTP_TOKENINFO,
                                    chl->otp_tokenInfo.list.array[i], 0);
  if (chl->otp_tokenInfo.list.count > i + 1)
    memmove(chl->otp_tokenInfo.list.array + i,
            chl->otp_tokenInfo.list.array + i + 1,
            sizeof(OTP_TOKENINFO_t*) * (chl->otp_tokenInfo.list.count - i - 1));
  chl->otp_tokenInfo.list.count--;
}

static krb5_error_code
prompt(krb5_context context, krb5_prompter_fct prompter, void *prompter_data,
       const char *banner, const char *prompttxt, char *out, size_t len)
{
  krb5_prompt prompt;
  krb5_data prompt_reply;
  krb5_error_code retval;

  if (!context || !prompter || !prompttxt || !out)
    return EINVAL;

  memset(out, 0, len);

  prompt_reply.data = out;
  prompt_reply.length = len;
  prompt.reply = &prompt_reply;
  prompt.prompt = (char*) prompttxt;
  prompt.hidden = 1;

  retval = (*prompter)(context, prompter_data, NULL, banner, 1, &prompt);
  if (retval) {
    DEBUGMSG(retval, "Unable to prompt!");
    return retval;
  }

  return 0;
}

static krb5_error_code
choose_token(krb5_context context, krb5_prompter_fct prompter,
             void *prompter_data, PA_OTP_CHALLENGE_t *chl)
{
  int i = 0, j = 0;
  char *banner = NULL, *tmp = NULL, tmpa[1024];
  krb5_error_code retval;
  OTP_TOKENINFO_t *ti;

  for (i = 0; i < chl->otp_tokenInfo.list.count; i++) {
    ti = chl->otp_tokenInfo.list.array[i];

    if (asprintf(&tmp, "%s\t%d. Vendor: %s\n",
                 banner ? banner : "Please choose from the following:\n",
                 i, ti->otp_vendor ? ti->otp_vendor->buf : NULL) < 0) {
      free(banner);
      return ENOMEM;
    }

    free(banner);
    banner = tmp;
  }

  while (true) {
    retval = prompt(context, prompter, prompter_data, banner,
                    "Enter #", tmpa, sizeof(tmpa));
    if (retval) {
      free(banner);
      return retval;
    }

    if (sscanf(tmpa, "%d", &j) != 1)
      continue;

    j--;
    if (j < 0 || j > chl->otp_tokenInfo.list.count - 1)
      continue;

    ti = chl->otp_tokenInfo.list.array[j];
    for (i = 0; i < chl->otp_tokenInfo.list.count; i++)
      if (chl->otp_tokenInfo.list.array[i] != ti)
        remove_tokeninfo(chl, i--);

    free(banner);
    return 0;
  }
}

static krb5_error_code
fill_in_request(krb5_context context, krb5_prompter_fct prompter,
                void *prompter_data, PA_OTP_CHALLENGE_t *chl,
                PA_OTP_REQUEST_t *req)
{
  krb5_error_code retval;
  int i, challengers = 0;
  char *challenge = NULL;
  const char *ptxt = NULL;
  OTP_TOKENINFO_t *ti = NULL;
  char zeroa(otpvalue, 1024);
  char zeroa(otppin, 1024);
  otpFlags supported_flags = otpFlagCollectPIN
                              | otpFlagDoNotCollectPIN
                              | otpFlagSeparatePINRequired;

  if (!chl || !req || chl->otp_tokenInfo.list.count <= 0)
    return EINVAL;

  /* Filter out tokeninfos that contain features we don't support */
  for (i = 0; i < chl->otp_tokenInfo.list.count; i++) {
    ti = chl->otp_tokenInfo.list.array[i];

    /* Flags we don't support... */
    if (int_asn1c_get_flags(&ti->flags) & ~supported_flags)
      goto remove;

    /* We don't currently support hashing */
    if (ti->supportedHashAlg || ti->iterationCount)
      goto remove;

    if (ti->otp_challenge)
      challengers++; /* Count how many challenges we have */

    continue;

  remove:
    remove_tokeninfo(chl, i--);
  }
  if (chl->otp_tokenInfo.list.count <= 0)
    return KRB5_PREAUTH_FAILED; /* We have no supported tokeninfos */

  /* Setup our challenge if present */
  if (challengers > 0) {
    /* If we have multiple tokeninfos still, choose now */
    if (chl->otp_tokenInfo.list.count > 1) {
      retval = choose_token(context, prompter, prompter_data, chl);
      if (retval)
        return retval;
    }

    /* Create the challenge prompt */
    if (chl->otp_tokenInfo.list.count > 0
        && chl->otp_tokenInfo.list.array[0]->otp_challenge) {
      if (asprintf(&challenge, "OTP Challenge: %s\n",
          chl->otp_tokenInfo.list.array[0]->otp_challenge->buf) < 0)
        return ENOMEM;
    }
  }

  /* Prompt for token value */
  ptxt = "Enter Password";
  for (i = 0; i < chl->otp_tokenInfo.list.count && ptxt; i++) {
    long int fmt;

    if (!chl->otp_tokenInfo.list.array[i]->otp_format)
      goto token;

    if (asn_INTEGER2long(chl->otp_tokenInfo.list.array[i]->otp_format,
                         &fmt) == 0)
      goto token;

    if (fmt != OTPFormat_alphanumeric)
      goto token;

    continue;

  token:
    ptxt = NULL;
  }
  retval = prompt(context, prompter, prompter_data, challenge,
                  ptxt ? ptxt : "Enter OTP Token Value",
                  otpvalue, sizeof(otpvalue));
  free(challenge);
  if (retval)
    return retval;

  /* Filter out tokeninfos that don't match our token value */
  for (i = 0; i < chl->otp_tokenInfo.list.count; i++)
    if (!token_matches_tokeninfo(otpvalue, chl->otp_tokenInfo.list.array[i]))
      remove_tokeninfo(chl, i--);

  /* If we still have multiplle tokeninfos, choose now */
  if (chl->otp_tokenInfo.list.count > 1) {
    retval = choose_token(context, prompter, prompter_data, chl);
    if (retval)
      return retval;
  }
  if (chl->otp_tokenInfo.list.count <= 0)
    return KRB5_PREAUTH_FAILED; /* We have no supported tokeninfos */

  /* We now have only one tokeninfo */
  ti = chl->otp_tokenInfo.list.array[0];

  /* Collect the PIN if necessary */
  if (int_asn1c_get_flags(&ti->flags) & otpFlagCollectPIN
      || int_asn1c_get_flags(&ti->flags) & otpFlagSeparatePINRequired) {
    retval = prompt(context, prompter, prompter_data, NULL, "OTP Token PIN",
                    otppin, sizeof(otppin));
    if (retval)
      return retval;

    /* Set the separate PIN and Value fields */
    if (int_asn1c_get_flags(&ti->flags) & otpFlagSeparatePINRequired) {
      asn_DEF_OCTET_STRING.free_struct(&asn_DEF_OCTET_STRING, req->otp_pin, 0);
      req->otp_pin = new0(UTF8String_t);
      if (OCTET_STRING_fromBuf(req->otp_pin, otppin, -1) < 0)
        return ENOMEM;

      asn_DEF_OCTET_STRING.free_struct(&asn_DEF_OCTET_STRING,
                                       req->otp_value, 0);
      req->otp_value = new0(OCTET_STRING_t);
      if (OCTET_STRING_fromBuf(req->otp_value, otpvalue, -1) < 0)
        return ENOMEM;

      /* Prepend PIN to the Value field */
    } else {
      char *tmp;

      if (asprintf(&tmp, "%s%s", otppin, otpvalue) < 0)
        return ENOMEM;

      asn_DEF_OCTET_STRING.free_struct(&asn_DEF_OCTET_STRING,
                                       req->otp_value, 0);
      req->otp_value = new0(OCTET_STRING_t);
      if (OCTET_STRING_fromBuf(req->otp_value, tmp, -1) < 0) {
        free(tmp);
        return ENOMEM;
      }
      free(tmp);
    }
  } else {
    asn_DEF_OCTET_STRING.free_struct(&asn_DEF_OCTET_STRING, req->otp_value, 0);
    req->otp_value = new0(OCTET_STRING_t);
    if (OCTET_STRING_fromBuf(req->otp_value, otpvalue, -1) < 0)
      return ENOMEM;
  }

  /* Copy values in from the tokeinfo */
  asn_DEF_OTPFlags.free_struct(&asn_DEF_OTPFlags, &req->flags, 1);
  req->flags = ti->flags;
  memset(&ti->flags, 0, sizeof(ti->flags));
  COPY_FIELD(req, ti, otp_algID, AnyURI);
  COPY_FIELD(req, ti, otp_format, OTPFormat);
  COPY_FIELD(req, ti, otp_tokenID, OCTET_STRING);
  COPY_FIELD(req, ti, otp_vendor, UTF8String);

  return 0;
}

static krb5_error_code
authhub_client_process(krb5_context context, krb5_clpreauth_moddata moddata,
                       krb5_clpreauth_modreq modreq,
                       krb5_get_init_creds_opt *opt,
                       krb5_clpreauth_callbacks cb, krb5_clpreauth_rock rock,
                       krb5_kdc_req *request, krb5_data *encoded_request_body,
                       krb5_data *encoded_previous_request,
                       krb5_pa_data *pa_data, krb5_prompter_fct prompter,
                       void *prompter_data, krb5_pa_data ***pa_data_out)
{
  struct otp_client_ctx *otp_ctx = (struct otp_client_ctx *) moddata;

  krb5_error_code retval = 0;
  krb5_keyblock *as_key = NULL;
  PA_OTP_REQUEST_t *req = (PA_OTP_REQUEST_t*) moddata;

  PA_OTP_CHALLENGE_t *chl = NULL;

  /*
   * Check for sanity
   */
  if (!otp_ctx) {
    DEBUGMSG(EINVAL, "Missing request.");
    return EINVAL;
  }

  if (pa_data->pa_type != KRB5_PADATA_OTP_CHALLENGE) {
    DEBUGMSG(ENOENT, "Unexpected PA data.");
    return ENOENT;
  }

  if (pa_data->length <= 0) {
    DEBUGMSG(ENOENT, "Empty PA data.");
    return ENOENT;
  }

  /* Get FAST armor key */
  as_key = cb->fast_armor(context, rock);
  if (as_key == NULL)
  {
    DEBUGMSG(ENOENT, "Missing armor key.");
    return ENOENT;
  }

  /* Use FAST armor key as response key. */
  retval = cb->set_as_key(context, rock, as_key);
  if (retval != 0) {
    DEBUGMSG(retval, "Unable to set reply key.");
    return retval;
  }

  /* Decode the challenge */
  chl = int_ber_decode_(&asn_DEF_PA_OTP_CHALLENGE,
                        (const char*) pa_data->contents,
                        pa_data->length);
  if (!chl) {
    DEBUGMSG(EINVAL, "Unable to decode OTP challenge.");
    return EINVAL;
  }

  /* Fill in the request info from the TokenInfo structs */
  retval = fill_in_request(context, prompter, prompter_data, chl, req);
  if (retval) {
    asn_DEF_PA_OTP_CHALLENGE.free_struct(&asn_DEF_PA_OTP_CHALLENGE, chl, 0);
    return retval;
  }

  /* Encrypt the challenge's nonce and set it in the request */
  retval = encrypt_nonce(context, as_key, chl, req);
  asn_DEF_PA_OTP_CHALLENGE.free_struct(&asn_DEF_PA_OTP_CHALLENGE, chl, 0);
  if (retval) {
    DEBUGMSG(retval, "Unable to encrypt nonce.");
    return retval;
  }

  /* Verify our constraints */
  if (asn_DEF_PA_OTP_REQUEST.check_constraints(&asn_DEF_PA_OTP_REQUEST, req,
                                               on_constraint_fail, NULL) < 0) {
    DEBUGMSG(retval, "Unable to create well-formed request.");
    return retval;
  }

  /* Allocate the preauth data array and one item */
  *pa_data_out = calloc(2, sizeof(krb5_pa_data *));
  if (*pa_data_out == NULL)
    return ENOMEM;
  (*pa_data_out)[0] = calloc(1, sizeof(krb5_pa_data));
  (*pa_data_out)[1] = NULL;
  if ((*pa_data_out)[0] == NULL) {
    free(*pa_data_out);
    return ENOMEM;
  }

  /* Encode our request into the preauth data item */
  memset((*pa_data_out)[0], 0, sizeof(krb5_pa_data));
  (*pa_data_out)[0]->pa_type = KRB5_PADATA_OTP_REQUEST;
  if (!int_der_encode(&asn_DEF_PA_OTP_REQUEST, req,
      (char **) &(*pa_data_out)[0]->contents, &(*pa_data_out)[0]->length)) {
    DEBUGMSG(ENOMEM, "Request encoding failed.");
    free((*pa_data_out)[0]);
    free(*pa_data_out);
    return ENOMEM;
  }

#ifdef DEBUG
  char *debug = NULL;
  if (int_asn1c_dump(&asn_DEF_PA_OTP_REQUEST, req, &debug)) {
    DEBUGMSG(0, "\n%s\n\n", debug);
    free(debug);
  }
#endif

  return 0;
}

krb5_error_code
clpreauth_authhub_initvt(krb5_context context, int maj_ver, int min_ver,
                         krb5_plugin_vtable vtable)
{
  krb5_clpreauth_vtable vt;

  if (maj_ver != 1)
    return KRB5_PLUGIN_VER_NOTSUPP;

  vt = (krb5_clpreauth_vtable) vtable;
  vt->name = "authhub";
  vt->pa_type_list = authhub_client_supported_pa_types;
  vt->init = authhub_client_init;
  vt->fini = authhub_client_fini;
  vt->flags = authhub_client_get_flags;
  vt->process = authhub_client_process;
  vt->gic_opts = NULL;

  DEBUGMSG(0, "Loaded.");
  return 0;
}
