/*
 * Copyright 2011 Red Hat, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "plugins.h"
#include "memory.h"
#include "internal.h"

#include "codec_json.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

typedef struct plugin   plugin;
typedef struct ti_state ti_state;
typedef struct ti_misc  ti_misc;
typedef struct vr_misc  vr_misc;

struct plugin {
  inventory   *pinv;
  plugin      *next;
  verto_ev    *jrpc;
  char        *name; /* basename */
  char        *file; /* abspath */
  int          sock;
};

struct inventory {
  plugin *plugins;
  struct {
    struct {
      char   *name;
      plugin *plug;
    } *mtbl;
    size_t  size;
    size_t  used;
  } map;
};

struct ti_state {
  size_t ref;
  PA_OTP_CHALLENGE_t *chl;
  void (*respond)(PA_OTP_CHALLENGE_t *chl, void *misc);
  jsonrpc_free *mfree;
  void *misc;
};

struct ti_misc {
  plugin      *plug;
  json_object *conf;
  ti_state    *stat;
};

struct vr_misc {
  void (*respond)(bool verified, void *misc);
  jsonrpc_free *mfree;
  void *misc;
};

static void
plugin_free(plugin *p)
{
  if (p) {
    free(p->name);
    free(p->file);
    if (p->sock >= 0)
      close(p->sock);
  }
  free(p);
}

static void
ti_state_decref(ti_state *tistate)
{
  if (!tistate)
    return;

  tistate->ref = tistate->ref > 1 ? tistate->ref - 1 : 0;
  if (tistate->ref == 0) {
    tistate->respond(tistate->chl, tistate->misc);
    if (tistate->mfree)
      tistate->mfree(tistate->misc);
    free(tistate);
  }
}

static void
ti_misc_free(ti_misc *timisc)
{
  if (!timisc)
    return;

  json_object_put(timisc->conf);
  ti_state_decref(timisc->stat);
  free(timisc);
}

static void
vr_misc_free(vr_misc *vrmisc)
{
  if (!vrmisc)
    return;

  if (vrmisc->mfree)
      vrmisc->mfree(vrmisc->misc);
  free(vrmisc);
}

static plugin *
plugin_map_get(inventory *inv, const char *name)
{
  if (!inv || !name)
    return NULL;

  for (int i=0 ; i < inv->map.used ; i++)
    if (!strcmp(inv->map.mtbl[i].name, name))
      return inv->map.mtbl[i].plug;

  return NULL;
}

static int
plugin_map_set(inventory *inv, const char *name, int len, plugin *plug)
{
  void *tmp;

  /* Update it if it exists */
  for (int i=0 ; i < inv->map.used ; i++) {
    if (!strcmp(inv->map.mtbl[i].name, name)) {
      inv->map.mtbl[i].plug = plug;
      return 0;
    }
  }

  /* Resize if necessary */
  if (inv->map.size == inv->map.used * sizeof(*inv->map.mtbl)) {
    tmp = realloc(inv->map.mtbl, inv->map.size ? inv->map.size * 2
                                               : sizeof(*inv->map.mtbl));
    if (!tmp)
      return ENOMEM;
    inv->map.mtbl = tmp;
    inv->map.size = inv->map.size ? inv->map.size * 2 : sizeof(*inv->map.mtbl);
  }

  /* Set the name */
  inv->map.mtbl[inv->map.used].name = strndup(name, len);
  if (!inv->map.mtbl[inv->map.used].name)
    return ENOMEM;

  /* Set the plugin */
  inv->map.mtbl[inv->map.used++].plug = plug;
  return 0;
}

static int
plugin_connect(plugin *plug)
{
  struct sockaddr_un addr;
  struct stat st;
  int retval;

  if (!plug || !plug->file)
    return EINVAL;

  if (stat(plug->file, &st) < 0) {
    retval = errno;
    DEBUGMSG(errno, "Unable to stat %s!", plug->file);
    return retval;
  }

  if (!S_ISSOCK(st.st_mode)) {
    DEBUGMSG(ENOENT, "%s is not a socket!", plug->file);
    return ENOENT;
  }

  if (st.st_uid != 0 || st.st_mode & (S_IRWXG | S_IRWXO)) {
    DEBUGMSG(EPERM, "%s has bad permissions!", plug->file);
    return EPERM;
  }

  plug->sock = socket(AF_UNIX, SOCK_STREAM, 0);
  if (plug->sock < 0) {
    retval = errno;
    DEBUGMSG(errno, "Unable to create UNIX socket!");
    return retval;
  }

  addr.sun_family = AF_UNIX;
  strncpy(addr.sun_path, plug->file, sizeof(addr.sun_path));
  if (connect(plug->sock, &addr, sizeof(addr)) < 0) {
    retval = errno;
    DEBUGMSG(errno, "Unable to connect to %s!", plug->file);
    close(plug->sock);
    return retval;
  }

  return 0;
}

static void
plugin_close(inventory *self, plugin *plug)
{

  for (int i=0 ; i < self->map.used ; i++) {
    if (self->map.mtbl[i].plug == plug) {
      free(self->map.mtbl[i].name);
      self->map.used--;
      if (i < self->map.used)
        memmove(&self->map.mtbl[i], &self->map.mtbl[i+1],
                self->map.used - i * sizeof(*self->map.mtbl));
    }
  }

  for (plugin *l = NULL, *p = self->plugins ; p ; l = p, p = p->next) {
    if (p == plug) {
      if (!l)
        self->plugins = p->next;
      else
        l->next = p->next;

      plugin_free(p);
      break;
    }
  }
}

static void
err_cb(verto_ev *ev, int errnum, void *misc)
{
  inventory *self = (inventory*) misc;

  for (plugin *p = self->plugins ; p ; p = p->next) {
    if (p->jrpc == ev) {
      DEBUGMSG(errnum, "Closing plugin: %s", p->name);
      p->jrpc = NULL;

      errno = plugin_connect(p);
      if (errno != 0)
        plugin_close(self, p);
      else
        DEBUGMSG(0, "Re-opened plugin: %s", p->name);

      break;
    }
  }

  verto_del(ev);
}

int
plugins_load(const char *directory, inventory **self)
{
  DIR *d;
  struct dirent *dent;
  struct stat st;

  if (stat(directory, &st) != 0) {
    DEBUGMSG(errno, "Unable to stat directory %s!", directory);
    return errno;
  }

  if (!S_ISDIR(st.st_mode)) {
    DEBUGMSG(ENOTDIR, "%s is not a directory!", directory);
    return ENOTDIR;
  }

  if (st.st_uid != 0 || st.st_mode & (S_IWGRP | S_IWOTH)) {
    DEBUGMSG(EPERM, "%s has bad permissions!", directory);
    return EPERM;
  }

  d = opendir(directory);
  if (!d) {
    DEBUGMSG(errno, "Unable to open %s!", directory);
    return errno;
  }

  *self = new0(inventory);
  if (!*self) {
    closedir(d);
    return ENOMEM;
  }

  while ((dent = readdir(d))) {
    plugin *plug;

    if (dent->d_name[0] == '.')
      continue;

    plug = new0(plugin);
    if (!plug) {
      closedir(d);
      plugins_free(*self);
      return ENOMEM;
    }

    plug->name = strdup(dent->d_name);
    if (!plug->name) {
      closedir(d);
      plugins_free(*self);
      plugin_free(plug);
      return ENOMEM;
    }

    if (asprintf(&plug->file, "%s/%s", directory, dent->d_name) < 0) {
      closedir(d);
      plugins_free(*self);
      plugin_free(plug);
      return ENOMEM;
    }

    errno = plugin_connect(plug);
    if (errno != 0) {
      DEBUGMSG(errno, "Failed to connect to plugin: %s", plug->name);
      plugin_free(plug);
      continue;
    }

    DEBUGMSG(0, "Registered plugin: %s", plug->name);

    plug->next = (*self)->plugins;
    (*self)->plugins = plug;
    plug->pinv = *self;
  }

  closedir(d);

  if (!(*self)->plugins) {
    DEBUGMSG(ENOENT, "No plugins found!");
    plugins_free(*self);
    return ENOENT;
  }

  return 0;
}

static void
on_ti_response(json_object *response, bool error, void *misc)
{
  ti_misc *timisc = (ti_misc*) misc;
  PA_OTP_CHALLENGE_t *chl = timisc->stat->chl;
  void *tmp;
  int len;

  /* The response must be an array */
  if (error) {
    DEBUGMSG(EINVAL, "Response is an error.");
    return;
  }

  if (!json_object_is_type(response, json_type_array)) {
    DEBUGMSG(EINVAL, "Response is not an array.");
    return;
  }

  /* Figure out how many token infos we got back */
  len = json_object_array_length(response);
  if (len < 1) {
    DEBUGMSG(EINVAL, "Response has no token infos.");
    return;
  }

  /* Resize the array */
  tmp = realloc(chl->otp_tokenInfo.list.array,
                sizeof(OTP_TOKENINFO_t*)
                  * (chl->otp_tokenInfo.list.count + len));
  if (!tmp)
    return;
  chl->otp_tokenInfo.list.array = tmp;

  /* Append the tokeninfos */
  for (int i=0 ; i < len ; i++) {
    json_object *obj;
    OTP_TOKENINFO_t *ti;

    obj = json_object_array_get_idx(response, i);
    if (!obj)
      continue;

    ti = new0(OTP_TOKENINFO_t);
    if (!ti)
      continue;

    if (!int_json_decode_object(&asn_DEF_OTP_TOKENINFO, ti, obj)) {
      DEBUGMSG(EINVAL, "Unable to decode token info.");
      asn_DEF_OTP_TOKENINFO.free_struct(&asn_DEF_OTP_TOKENINFO, ti, 0);
      continue;
    }

    /* The vendor field is REQUIRED */
    if (!ti->otp_vendor || ti->otp_vendor->size == 0) {
      asn_DEF_OTP_TOKENINFO.free_struct(&asn_DEF_OTP_TOKENINFO, ti, 0);
      continue;
    }

    /* Add this vendor to our map */
    plugin_map_set(timisc->plug->pinv, (char*) ti->otp_vendor->buf,
                   ti->otp_vendor->size, timisc->plug);

    chl->otp_tokenInfo.list.array[chl->otp_tokenInfo.list.count++] = ti;
  }
}

int
plugins_getTokenInfo(inventory *self, verto_ctx *ctx, PA_OTP_CHALLENGE_t *chl,
                     json_object *(*config)(const char *name, const void *misc),
                     void (*respond)(PA_OTP_CHALLENGE_t *chl, void *misc),
                     const void *cmisc, void *rmisc, jsonrpc_free *rmisc_free)
{
  json_object *params;
  ti_misc **timisc;
  ti_state *tistate;
  int count, i;

  if (!self || !ctx || !chl || !config || !respond) {
    if (rmisc_free)
      rmisc_free(rmisc);
    return EINVAL;
  }

  /* Make sure all our jrpcs are setup */
  count = 0;
  for (plugin *p = self->plugins ; p ; p = p->next) {
    if (!p->jrpc) {
      p->jrpc = jsonrpc_init(ctx, p->sock, p->sock, 1024 * 512,
                             err_cb, self, NULL);
      if (!p->jrpc) {
        DEBUGMSG(ENOMEM, "Unable to create JSON-RPC event.");
        if (rmisc_free)
          rmisc_free(rmisc);
        return ENOMEM;
      }
    }

    count++;
  }

  /* Build the params object */
  params = json_object_new_object();
  if (!params) {
    if (rmisc_free)
      rmisc_free(rmisc);
    return ENOMEM;
  }

  /* Build the respond state */
  tistate = new0(ti_state);
  if (!tistate) {
    json_object_put(params);
    if (rmisc_free)
      rmisc_free(rmisc);
    return ENOMEM;
  }
  tistate->mfree = rmisc_free;
  tistate->misc = rmisc;
  tistate->respond = respond;
  tistate->chl = chl;

  /* Build our temporary ti_misc array */
  timisc = newa0(ti_misc*, count + 1);
  if (!timisc) {
    json_object_put(params);
    ti_state_decref(tistate);
    return ENOMEM;
  }

  /* Create one ti_misc for each plugin with config */
  i = 0;
  for (plugin *p = self->plugins ; p && i < count ; p = p->next) {
    timisc[i] = new0(ti_misc);
    if (!timisc[i]) {
      if (i == 0)
        ti_state_decref(tistate);
      for (--i ; i >= 0 ; i--)
        ti_misc_free(timisc[i]);
      json_object_put(params);
      return ENOMEM;
    }

    timisc[i]->conf = config(p->name, cmisc);
    if (!timisc[i]->conf) {
      free(timisc[i]);
      timisc[i] = NULL;
      continue;
    }

    timisc[i++]->stat = tistate;
    timisc[i-1]->plug = p;
    tistate->ref++;
  }

  /* Do the call */
  for (i=0 ; timisc[i] ; i++) {
    json_object_object_add(params, "config", json_object_get(timisc[i]->conf));
    jsonrpc_call(timisc[i]->plug->jrpc, NULL, "authhub::getTokenInfo",
                 json_object_get(params), on_ti_response,
                 timisc[i], (jsonrpc_free*) ti_misc_free);
  }

  free(timisc);
  json_object_put(params);
  return i > 0 ? 0 : ENOENT; /* If no plugins have config, return an error */
}

static void
on_vr_response(json_object *response, bool error, void *misc)
{
  vr_misc *vrmisc = (vr_misc*) misc;

  if (!error && json_object_is_type(response, json_type_boolean)
      && json_object_get_boolean(response))
    vrmisc->respond(true, vrmisc->misc);
  else
    vrmisc->respond(false, vrmisc->misc);
}

int
plugins_verifyRequest(inventory *self, verto_ctx *ctx, PA_OTP_REQUEST_t *req,
                      json_object *(*config)(const char *name, const void *misc),
                      void (*respond)(bool success, void *misc),
                      const void *cmisc, void *rmisc, jsonrpc_free *rmisc_free)
{
  json_object *params, *vendor, *cnfg, *jreq;
  int count;
  const char *vendstr;
  plugin *plug;
  vr_misc *vrmisc;

  if (!self || !ctx || !req || !config || !respond) {
    if (rmisc_free)
      rmisc_free(rmisc);
    return EINVAL;
  }

  /* Make sure all our jrpcs are setup */
  count = 0;
  for (plugin *p = self->plugins ; p ; p = p->next) {
    if (!p->jrpc) {
      p->jrpc = jsonrpc_init(ctx, p->sock, p->sock, 1024 * 512,
                             err_cb, self, NULL);
      if (!p->jrpc) {
        DEBUGMSG(ENOMEM, "Unable to create JSON-RPC event.");
        if (rmisc_free)
          rmisc_free(rmisc);
        return ENOMEM;
      }
    }

    count++;
  }

  /* Encode the request to json */
  jreq = int_json_encode_object(&asn_DEF_PA_OTP_REQUEST, req);
  if (!jreq) {
    DEBUGMSG(ENOMEM, "Unable to encode request.");
    if (rmisc_free)
      rmisc_free(rmisc);
    return ENOMEM;
  }

  /* Allocate vr_misc */
  vrmisc = new0(vr_misc);
  if (!vrmisc) {
    if (rmisc_free)
      rmisc_free(rmisc);
    json_object_put(jreq);
    return ENOMEM;
  }
  vrmisc->mfree = rmisc_free;
  vrmisc->respond = respond;
  vrmisc->misc = rmisc;

  /* Get the vendor requested */
  vendor = json_object_object_get(jreq, "otp-vendor");
  if (!vendor || !json_object_is_type(vendor, json_type_string)
      || !(vendstr = json_object_get_string(vendor))) {
    DEBUGMSG(EINVAL, "Request MUST include a vendor.");
    vr_misc_free(vrmisc);
    json_object_put(jreq);
    return EINVAL;
  }

  /* Get the plugin for this vendor */
  plug = plugin_map_get(self, vendstr);
  if (!plug) {
    DEBUGMSG(EINVAL, "No plugin found matching vendor '%s'.", vendstr);
    vr_misc_free(vrmisc);
    json_object_put(jreq);
    return EINVAL;
  }

  /* Get the configuration for the plugin of the vendor requested */
  cnfg = config(plug->name, cmisc);
  if (!cnfg) {
    DEBUGMSG(EINVAL, "User is not configured for requested vendor '%s.", vendstr);
    vr_misc_free(vrmisc);
    json_object_put(jreq);
    return EINVAL;
  }

  /* Build the params object */
  params = json_object_new_object();
  if (!params) {
    vr_misc_free(vrmisc);
    json_object_put(jreq);
    json_object_put(cnfg);
    return ENOMEM;
  }
  json_object_object_add(params, "request", jreq);
  json_object_object_add(params, "config",  cnfg);

  /* Do the call */
  if (!jsonrpc_call(plug->jrpc, NULL, "authhub::verifyRequest", params,
                    on_vr_response, vrmisc, (jsonrpc_free*) vr_misc_free))
    return ENOMEM;

  return 0;
}

void
plugins_free(inventory *self)
{
  plugin *l = NULL;

  for (int i = 0 ; i < self->map.used ; i++)
    free(self->map.mtbl[i].name);
  for (plugin *p = self->plugins ; p ; l = p, p = p->next)
    plugin_free(l);
  plugin_free(l);
  free(self);
}
